(ns com.aurellem.assembly
  (:use (com.aurellem gb-driver vbm title items))
  (:import [com.aurellem.gb_driver SaveState]))

(defn mid-game []
  (read-state "mid-game"))

(defn inject-assembly
  ([^SaveState state
   program-counter registers
   assembly-code]
  (let [scratch-memory (memory state)]
    ;; inject assembly code
    (dorun (map (fn [index val]
                  (aset scratch-memory index val))
                (range program-counter
                       (+ program-counter (count assembly-code)))
                assembly-code))
    (-> state
        (write-memory! scratch-memory)
        (write-registers! registers)
        (PC! program-counter))))
  ;;([program-counter]
  ;;   (fn [^SaveState state registers assembly-code]
  ;;     (inject-assembly state program-counter registers assembly-code)))
  )


;;(def inject-assembly-item
;;  (inject-assembly (inc item-list-start)))

(defn inject-item-assembly
  ([^SaveState state assembly-code]
     (inject-assembly state (inc item-list-start)
                      (registers state)
                      assembly-code))
  ([assembly-code]
     (inject-item-assembly @current-state assembly-code)))

(defn info
  ([^SaveState state]
     (println (format "PC: 0x%04X" (PC state)))
     (println "Instruction:"
              (format "0x%02X" (aget (memory state) (PC state))))
     state))    

(defn print-interrupt
  [^SaveState state]
  (println (format "IE: %d" (IE state)))
  state)

(defn run-assembly
  ([info-fn assembly n]
     (let [final-state
           (reduce (fn [state _]
                     (tick (info-fn state)))
                   (inject-item-assembly
                     (mid-game) assembly)
                   (range n))]
       final-state))
  ([assembly n]
     (run-assembly info assembly n)))


(def buttons-port 0xFF00)

(defn view-register [state name reg-fn]
  (println (format "%s : 0x%02X" name (reg-fn state)))
  state)

(defn A [state]
  (bit-shift-right (bit-and 0x0000FF00 (AF state)) 8))

(defn view-memory [state mem]
  (println (format "mem 0x%04X = 0x%02X" mem (aget (memory state) mem)))
  state)

(defn read-buttons []
  (let [button-pressed (tick (step (mid-game) [:d]))]
    (-> button-pressed
        (IE! 0) ; disable interrupts
        (inject-item-assembly
         (concat
          ;; write 00010000 to 0xFF00 to select joypad
          [0x18   ;D31D                 ; jump over          
           0x01   ;D31E                 ; the next 8 bits
           (Integer/parseInt "00010000" 2) ;D31F data section 
           0x00   ;D320                 ; take a break
           0xFA   ;D321                 ; put DC15 into A
           0x1F   ;D322      --]       
           0xD3   ;D323      --]  DC1F  ; data section
           0x01   ;D324                 ; load 0xFF00 into BC
           0x00   ;D325
           0xFF   ;D326
           0x02   ;D327    (BC)->A      ; load (00010000) into FF00
                                        ; to select directional
                                        ; buttons

           ]

          []))
        (info)
        (tick)  ;; skip over data section
        (info)  
        (tick)  ;; no-op
        (info)
        (view-register "Register A" A)
        (tick)  ;; load-data into A
        (view-register "Register A" A)
        (info)
        (view-register "Register BC" BC)
        (tick) ;;  store 0xFFOO into BC
        (view-register "Register BC" BC)
        (info)
        (view-memory 0xFF00)
        (tick) ;; load A into 0xFF00
        (view-memory 0xFF00)
        )))
          
         
(defn trace [state]
  (loop [program-counters []
         opcodes []]
    (let [frame-boundary?
          (com.aurellem.gb.Gb/tick)]
      (println (count opcodes))
      (if frame-boundary?
        [program-counters opcodes]
        (recur
         (conj program-counters
               (first (registers @current-state)))
         (conj opcodes
               (aget (memory @current-state)
                     (PC @current-state))))))))
        
  
(defn good-trace []
  (-> (mid-game) (tick) (IE! 0)
      (set-inv-mem [0x00 0x00 0X00 0x00])
      (PC! item-list-start)(print-interrupt)
      (info) (tick) (info) (tick) (info)))